/*
 *      vpatch - binary patcher based on 'as'.
 *      (as's ugly parameter parsing has been replaced with a more gnu-like syntax)
 *
 *   I built this code based on as9 from https://www.cs.drexel.edu/~bls96/6809sbc/tools/as9/
 *   and have converted some of the older C constructs into more modern C.  It should
 *   compile cleanly with no warnings by all modern C compilers.  - gtoal@gtoal.com 2021/12/21
 *
 *   Other versions of the same assembler are online at various projects:
 *
 *     http://cd.textfiles.com/geminiatari/FILES/LANGUAGE/MOTASMS/ (may be the original online copy)
 *     https://github.com/pdaxrom/Z-pdaXrom-TV/tree/master/projects/p601zero-plus/ROM/DEVEL/AS11
 *     https://github.com/JimInCA/motorola-6800-assembler/tree/master/src
 *     https://github.com/MattisLind/6801reader/tree/master/software/motasm_linux
 *     https://github.com/freecores/System09/tree/master/Tools/as09
 *     https://opencores.org/websvn/listing?repname=System09&path=%2FSystem09%2Ftrunk%2FTools%2Fas09%2F&rev=206#path_System09_trunk_Tools_as09_
 *
 *   There are also other assemblers based on the original which do not support the 6809.
 */

void help(void) {
  fprintf(stderr, "syntax: %s [options] bin_file [patch_files]\n", ProgName);
  fprintf(stderr, "\n");
  fprintf(stderr, "Patch a binary (rom) file with one or more assembly patches.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "   Long Options   Short Options\n");
  fprintf(stderr, "   --debug        -D\n");
  fprintf(stderr, "   --help         -h\n");
  fprintf(stderr, "   --[no]list     -l       Create a .lst output file\n");
  fprintf(stderr, "   --[no]cycles   -c       Include cycle count in listing\n");
  fprintf(stderr, "   --[no]symbols  -s       Generate a .sym symbox file\n");
  fprintf(stderr, "   --xref         -x       Create a .crf cross-reference file\n");
  fprintf(stderr, "   --bin          -b       Generate a binary image\n");
  fprintf(stderr, "   --s19          -S       Create a Motorola S-record output file\n");
  fprintf(stderr, "   --warn         -w       Print warning messages\n");
  fprintf(stderr, "   --quiet        -q       Suppress info messages\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Short options may be combined, e.g. -wlS\n");
  fprintf(stderr, "\n");
  exit(EXIT_SUCCESS);
}

void terminate(void) {
  /* Just in case someone calls this explicitly before exiting, this is a
     little defensive safeguard against calling it twice. */
  static int terminated = 0;
  if (terminated > 0) return;
  terminated = 1;
  /* still to do:
eval.c:        pnt = (struct link *) malloc (sizeof (struct link));
symtab.c:  np = (struct nlist *) malloc (sizeof (struct nlist));
symtab.c:  np->name = malloc (strlen (str) + 1);
symtab.c:  lp = (struct link *) malloc (sizeof (struct link));
   */
  fwdterminate ();
  localterminate ();
  if (!DEBUG) remove (Fwd_name);
  if (DEBUG) fprintf(stderr, "Cleaned up.\n");
}


int main (int argc, char **argv) {
  char **np;

  ProgName = argv[0]; /* save program name before shifting args */
  if (((ProgName = strrchr(argv[0], '/')) != NULL) ||
      ((ProgName = strrchr(argv[0], '\\')) != NULL) ||
      ((ProgName = strrchr(argv[0], ':')) != NULL)) {
    ProgName += 1; /* skip '/' or '\' or ':' */
  } else {
    ProgName = argv[0];
  }
  
  if (argc < 2) {
    fprintf (stderr, "Usage: %s [options] [files]   (try \"%s -h\" for more info)\n", ProgName, ProgName);
    exit (EXIT_FAILURE);
  }

  pre_initialize ();
  
  /* Option arguments first */
  while ((argc >= 2) && (argv[1][0] == '-')) {
    if (strcmp (argv[1], "--debug") == 0) {
      DEBUG = 1; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--help") == 0) {
      help();
    } else if (strcmp (argv[1], "--list") == 0) {
      Lflag = 1; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--nolist") == 0) {       /* append .lst to list file name. */
      Lflag = 0; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--cycles") == 0) {
      Cflag = 1; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--nocycles") == 0) {
      Cflag = 0; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "-symbols") == 0) {       /* append .sym to symbol file name. */
      Sflag = 1; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "-nosymbols") == 0) {     /* append .sym to symbol file name. */
      Sflag = 0; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--cref") == 0) {         /* append .crf to cross reference file name. */
      Rflag = 1; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--xref") == 0) {         /* append .crf to cross reference file name. */
      Rflag = 1; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--nocref") == 0) {       /* append .crf to cross reference file name. */
      Rflag = 0; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--noxref") == 0) {       /* append .crf to cross reference file name. */
      Rflag = 0; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--bin") == 0) {          /* append .bin to binary file name. */
      Bflag = 1; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--nobin") == 0) {        /* append .bin to binary file name. */
      Bflag = 0; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--s19") == 0) {          /* append .s19 to object file name. */
      Oflag = 1; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--nos19") == 0) {        /* append .s19 to object file name. */
      Oflag = 0; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--warn") == 0) {
      Wflag = 1; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--nowarn") == 0) {
      Wflag = 1; argc -= 1; argv += 1;
    } else if (strcmp (argv[1], "--warn") == 0) {
      Qflag = 1; argc -= 1; argv += 1;
    } else if (argv[1][1] == '-') {
      fatal ("Unknown long option \"%s\"", argv[1]);
    } else {
      char *argptr = argv[1]+1; /* skip the '-' */
      while (*argptr != '\0') {
        switch (*argptr) {
        case 'd':
          fatal ("Unknown option \"-%c\" - did you mean \"-D\" perhaps?", *argptr);
        case 'D':
          DEBUG = 1; break;
        case 'h':
          help();

        case 'b':
          Bflag = 1; break;
        case 'c':
        case 'x':
          Cflag = 1; break;
        case 'l':
          Lflag = 1; break;
        case 'q':
          Qflag = 1; break;
        case 's':
          Sflag = 1; break;
        case 'S':
          Oflag = 1; break;
        case 'w':
          Wflag = 1; break;
        default:
          fatal ("Unknown option \"-%c\"", *argptr);
          break;
        }
        argptr += 1;
      }
      argc -= 1; argv += 1;
    }
  }

  if (argc == 1) {
    fatal ("No files specified?");
  } else if (argc == 2) {
    fatal ("Need at least one bin file to patched and one patch file to apply to the bin.");
  }

  Patch_Binary = argv[1]; argc -= 1; argv += 1;

  { /* We need to check the header, but we have to load the whole rom anyway, so might as well do both */
    int i, c;
    Fd = fopen (Patch_Binary, "rb");
    if (Fd == NULL) {
      fatal ("Cannot open binary file \"%s\" to be patched - %s", ProgName, strerror(errno));
    }
    i = 0; /* All Vectrex ROMs start at 0000. We don't yet handle banked ROMs. */
    for (;;) {
      c = fgetc(Fd);
      if (c == EOF) break;
      mem[i] = c;
      i += 1;
    }
    ROM_Size = i;
    if (strncmp((char *)mem, "g GCE", 5)) {
      fatal ("The first file parameter should be a Vectrex ROM, but \"%s\" is not.", Patch_Binary);
    }
    fclose(Fd);

    info ("ROM_Size (from counting the bytes) is %d");
    ROM_CRC32 = crc32(mem, ROM_Size);
    debug ("CRC32 %s: %x", Patch_Binary, ROM_CRC32);
  }
  
  /* Positional file arguments are all that remain.  First should be bin file. Rest are patches. */
  N_files = argc-1;
  Argv = argv;

  if (DEBUG) {
    fprintf (stderr, "Initializing strings\n");
  }

  {
  char *dots;
  int dot;
  dots = strrchr(Patch_Binary, '.');
  if (dots == NULL) dot = strlen(Patch_Binary); else dot = dots-Patch_Binary;
  strcpy (Obj_name, Patch_Binary); Obj_name[dot] = '\0'; strcat(Obj_name, ".s19.tmp");   /* copy object file name into array */
  strcpy (Bin_name, Patch_Binary); Bin_name[dot] = '\0'; strcat(Bin_name, ".bin.tmp");   /* but don't override original      */
  strcpy (Lst_name, Patch_Binary); Lst_name[dot] = '\0'; strcat(Lst_name, ".lst");   /* copy list file name into array */
  strcpy (Sym_name, Patch_Binary); Sym_name[dot] = '\0'; strcat(Sym_name, ".sym");   /* copy symbol file name into array */
  strcpy (Crf_name, Patch_Binary); Crf_name[dot] = '\0'; strcat(Crf_name, ".crf");   /* copy crf file name into array */
  strcpy (Patched_output_name, "");
  }

  fwdinit ();                   /* forward ref init */
  localinit ();                 /* target machine specific init. */

  if (atexit(terminate) != 0) {
    fprintf(stderr, "cannot set exit function\n");
    exit(EXIT_FAILURE);
  }
  
  root = NULL;
  Cfn = 0;
  np = Argv;
  debug ("Pass %d", Pass);

  /* PASS 1 */
  while (++Cfn <= N_files) {
    Line_num = 0;                 /* reset line number */
    debug ("File %d", Cfn);
    if ((Fd = fopen (*++np, "r")) != NULL) {
      debug ("File %d make_pass", Cfn);
      make_pass ();
      debug ("File %d fclose", Cfn);
      fclose (Fd);
    } else {
      debug ("File %d error", Cfn);
      fprintf (stderr, "%s: can't open patch source %s - %s\n", ProgName, *np, strerror(errno)); /* why not just call fatal()? */
    }
    debug ("File %d end loop", Cfn);
  }
  debug ("Exited loop");
  if (Err_count != 0) {
    debug ("Err count %d", Err_count);
    exit (Err_count);
  }
  debug ("Next pass");
  Pass++;   /* Pass 2 */
  debug ("re_init()");
  re_init ();
  Cfn = 0;
  np = Argv;
  debug ("open_output_files()");
  open_output_files ();              /* Only write on second pass */

  debug ("Pass %d", Pass);
  /* PASS 2 */
  while (++Cfn <= N_files) {
    Line_num = 0;
    if ((Fd = fopen (*++np, "r")) != NULL) { make_pass (); fclose (Fd); }
    else fprintf (stderr, "%s: can't reopen patch source %s - %s\n", ProgName, *np, strerror(errno));
  }

  if (Sflag == 1) stable (root);
  if (Rflag == 1) cross (root);
  if (Oflag) fprintf (Objfil, "S9030000FC\n"); /* at least give a decent ending */

  /* Finally, dump mem[] to patched binary */
  {
    int rc;
    /*
        size_t fwrite(const void *ptr,
                      size_t size, size_t nmemb,
                      FILE *stream);
     */
    rc = fwrite(mem, 1, ROM_Size, Patchedfil);
    if (rc != ROM_Size) {
      fatal ("Unable to write entire output file \"%s\" only wrote %d bytes out of %d", rc, ROM_Size);
    }
    fclose(Patchedfil);
    if (!DEBUG) remove(Patch_name);
    info("Patch created successfully!");
  }

  exit (EXIT_SUCCESS);
  return EXIT_FAILURE;
}

void pre_initialize (void) {
  debug ("Initializing variables");
  Err_count = 0;
  for (Pc = 0; Pc <= 0xFFFF; Pc++) {
    mem[Pc] = 0xFF; /* if uninitialsed eproms are FF, you can overwrite them with your prommer... */
  }
  Pc = 0;
  Pass = 1;
  Lflag = 0;
  Cflag = 0;
  Ctotal = 0;
  Sflag = 0;
  Rflag = 0;
  Bflag = 0;
  Oflag = 0;
  N_page = 0;
}

void re_init (void) {
  if (DEBUG) {
    fprintf (stderr, "Reinitializing\n");
  }
  Pc = 0;
  E_total = 0;
  P_total = 0;
  Ctotal = 0;
  N_page = 0;
  fwdreinit ();
}

void open_output_files (void) {
  if ((Oflag) && ((Objfil = fopen (Obj_name, "w"))  == NULL)) fatal ("Can't create object file");
  if ((Bflag) && ((Binfil = fopen (Bin_name, "wb")) == NULL)) fatal ("Can't create binary file");
  if ((Lflag) && ((Lstfil = fopen (Lst_name, "w"))  == NULL)) fatal ("Can't create list file");
  if ((Sflag) && ((Symfil = fopen (Sym_name, "w"))  == NULL)) fatal ("Can't create symbol file");
  if ((Rflag) && ((Crffil = fopen (Crf_name, "w"))  == NULL)) fatal ("Can't create cross reference file");

  debug ("Output file name before adding patch name: %s", Patch_name);
  if (*Patched_output_name == '\0') strcpy(Patched_output_name, "-patched"); /* No-one used the PATCH directive */
  strcat(Patch_name, Patched_output_name); /* patch name includes a leading '-' */
  strcat(Patch_name, ".vec");
  info ("Writing patched file to \"%s\"", Patch_name);
  Patchedfil = fopen (Patch_name, "wb");
  if (Patchedfil == NULL) fatal ("Unable to open \"%s\" to save the patched ROM to.", Patch_name);
}

void make_pass (void) {
  int L;


  if (DEBUG) {
    fprintf (stderr, "Pass %d\n", Pass);
  }
  while (fgets (Line, MAXBUF, Fd)) {
    L = strlen (Line);
    if (L >= 2) {
      if ((Line[L-2] == '\r') && (Line[L-1] == '\n')) {
        Line[L-2] = Line[L-1];
        Line[L-1] = EOS;
        L -= 1;
      }
    }
    if ((L >= 1) && (Line[L-1] == '\n')) Line[--L] = EOS;
    Line[L] = '\n'; Line[L+1] = EOS; Line[L+2] = EOS;
    Line_num++;
    if (DEBUG) {
      fprintf (stderr, "Line %d: \"%s\" (len %d)\n", Line_num, Line, L);
    }
    P_force = 0;                /* No force unless bytes emitted */
    N_page = 0;
    if (parse_line ())
      process ();
    if (Pass == 2 && Lflag && !N_page)
      print_line ();
    P_total = 0;                /* reset byte count */
    Cycles = 0;                 /* and per instruction cycle count */
  }
  f_record ();
}

/*
 *      parse_line --- split input line into label, op and operand
 */
int parse_line (void) {
  register char *pcfrm = Line;
  register char *pcto;

  if (*pcfrm == '*' || *pcfrm == '\n' || *pcfrm == ';')
    return (0);                 /* a comment line */

  /***************************/

  pcto = Label;
  while (delim (*pcfrm) == NO)
    *pcto++ = *pcfrm++;
  /* allow trailing : */
  if (*pcfrm == ':')
    pcfrm++;
  *pcto = EOS;

  pcfrm = skip_white (pcfrm);

  /***************************/
  pcto = Op;
  while (delim (*pcfrm) == NO)
    *pcto++ = mapdn (*pcfrm++);
  *pcto = EOS;

  pcfrm = skip_white (pcfrm);

  /***************************/
  pcto = Operand;               /* Can contain comment */
  while (*pcfrm != NEWLINE && *pcfrm != EOS) {
    *pcto++ = *pcfrm++;
  }
  *pcto = EOS;

  if (DEBUG) {
    fprintf (stderr, "Label-%s-\n", Label);
    fprintf (stderr, "Op----%s-\n", Op);
    fprintf (stderr, "Operand-%s-\n", Operand);
  }
  
  return (1);
}

/*
 *      process --- determine mnemonic class and act on it
 */
void process (void) {
  register struct oper *i;

  Old_pc = Pc;                  /* setup `old' program counter */
  Optr = Operand;               /* point to beginning of operand field */

  if (*Op == EOS) {             /* no mnemonic */
    if (*Label != EOS)
      install (Label, Pc);
  } else if ((i = mne_look (Op)) == NULL)
    error ("Unrecognized Mnemonic \"%s\"", Op);
  else if (i->class == PSEUDO)
    do_pseudo (i->opcode);          /* plant data bytes */
  else {
    if (*Label)
      install (Label, Pc);
    if (Cflag)
      Cycles = i->cycles;
    do_op (i->opcode, i->class);     /* plant opcode */
    if (Cflag)
      Ctotal += Cycles;
  }

  if (Wflag && 2 == Pass)
    if (*Optr != EOS) {
      /* Yes I'm ashamed of this hack, (AH) */
      if (*Optr == ']')
        Optr += 1;
      Optr = skip_white (Optr);
      if (*Optr != EOS && *Optr != ';') {
        fprintf (stderr, "<<<%s>>>\n", Optr);
        warn ("Comment does not start with comment symbol");
      }
    }
}
