/*
 *      fatal --- fatal error handler
 */
void fatal (char *str, ...) {
  va_list ap;
  /* Any pass */
  
  fprintf (stderr, "%s: ", ProgName);

  va_start (ap, str);
  (void) vfprintf (stderr, str, ap);
  va_end (ap);
  fprintf (stderr, "\n\n");

  exit (EXIT_FAILURE);
}

/*
 *      error --- error in a line
 *                      print line number and error
 */
void error (char *str, ...) {
  va_list ap;

  if (Pass > 1) return; /* Does not increment Err_count on first pass */
  
  if ((N_files >= 1) && (Argv[Cfn] != NULL)) {
    fprintf (stderr, "\"%s\", ", Argv[Cfn]);
  }
  if (Line_num) fprintf (stderr, "Line %d: ", Line_num);
  fprintf (stderr, "Error: ");

  va_start (ap, str);
  (void) vfprintf (stderr, str, ap);
  va_end (ap);
  fprintf (stderr, "\n");
  Err_count++;
}

/*
 *      warn --- trivial error in a line
 *                      print line number and error
 */
void warn (char *str, ...) {
  va_list ap;

  if (Pass > 1) return;

  if ((N_files >= 1) && (Argv[Cfn] != NULL)) {
    fprintf (stderr, "\"%s\", ", Argv[Cfn]);
  }
  if (Line_num) fprintf (stderr, "Line %d: ", Line_num);
  fprintf (stderr, "Warning: ");
  va_start (ap, str);
  (void) vfprintf (stderr, str, ap);
  va_end (ap);
  fprintf (stderr, "\n");
}

void info (char *str, ...) {
  va_list ap;

  if ((Pass == 1) || Qflag) return;

  if ((N_files >= 1) && (Argv[Cfn] != NULL)) {
    fprintf (stderr, "\"%s\", ", Argv[Cfn]);
  }
  if (Line_num) fprintf (stderr, "Line %d: ", Line_num);
  va_start (ap, str);
  (void) vfprintf (stderr, str, ap);
  va_end (ap);
  fprintf (stderr, "\n");
}

void debug (char *str, ...) {
  va_list ap;

  if (!DEBUG) return;
  if ((N_files >= 1) && (Argv[Cfn] != NULL)) {
    fprintf (stderr, "\"%s\", ", Argv[Cfn]);
  }
  if (Line_num) fprintf (stderr, "Line %d: ", Line_num);
  va_start (ap, str);
  (void) vfprintf (stderr, str, ap);
  va_end (ap);
  fprintf (stderr, "\n");
}

/*
 *      delim --- check if character is a delimiter
 *      Slightly bizarre: !@#$%^&()_+ are all accepted
 *      as valid label characters.
 *      Blank and the comment sign are delimiters.
 */
int delim (char c) {
  if (c == '\0') return YES;
  return any (c, " :;\t\n");
}

/*
 *      skip_white --- move pointer to next non-whitespace char
 */
char *skip_white (char *ptr) {
  while (any (*ptr, " \n\t\r"))
    ptr++;
  return (ptr);
}

/*
 *      eword --- emit a word to code file
 */
void eword (int wd) {
  emit (hibyte (wd));
  emit (lobyte (wd));
}

/*
 *      emit --- emit a byte to code file
 */
void emit (int byte) {
  if (DEBUG) {
    fprintf (stderr, "%2x @ %4x\n", byte, Pc);
  }
  /* debug ("emit: Plant %02x at %04x", byte, Pc); // I'm not seeing this - would it be better to do in f_record instead? (flush record) */
  mem[Pc&0xFFFF] = byte&0xFF;
  if ((Pc&0xFFFF) > ROM_Size) {
    ROM_Size = Pc&0xFFFF; /* Extend the ROM to accomodate the new patched bytes which are after the previous end of file */
  }
  if (Pass == 1) {
    Pc++;
    return;
  }
  if (P_total < P_LIMIT)
    P_bytes[P_total++] = byte;
  E_bytes[E_total++] = byte;
  Pc++;
  if (E_total == E_LIMIT)
    f_record ();
}

/*
 *      f_record --- flush record out to S19 and Bin files if neccesary
 */
void f_record (void) {
  int i;
  int chksum;

  if (Pass == 1)
    return;
  if (E_total == 0) {
    E_pc = Pc;
    return;
  }
  chksum = E_total + 3;         /* total bytes in this record */
  chksum += lobyte (E_pc);
  chksum += E_pc >> 8;
  if (Oflag) {
    fprintf (Objfil, "S1");     /* record header preamble */
    hexout (E_total + 3);       /* byte count +3 */
    hexout (E_pc >> 8);         /* high byte of PC */
    hexout (lobyte (E_pc));     /* low byte of PC */
  }
  for (i = 0; i < E_total; i++) {
    chksum += lobyte (E_bytes[i]);
    if (Oflag)
      hexout (lobyte (E_bytes[i]));     /* data byte */
    if (Bflag)
      binout (E_bytes[i]);      /* binary data */
  }
  chksum = ~chksum;             /* one's complement */
  if (Oflag) {
    hexout (lobyte (chksum));   /* checksum */
    fprintf (Objfil, "\n");
  }
  E_pc = Pc;
  E_total = 0;
}

char *hexstr = "0123456789ABCDEF";

void hexout (int byte) {
  byte = lobyte (byte);
  fprintf (Objfil, "%c%c", hexstr[(byte >> 4) & 0xF], hexstr[byte & 0xF]);
}

void binout (int byte) {
  fprintf (Binfil, "%c", byte);
}

/*
 *      print_line --- pretty print input line to List file
 */
void print_line (void) {
  int i;

  if (Lflag == 0) return;

  fprintf (Lstfil, "%04d ", Line_num);

  if (P_total || P_force)
    fprintf (Lstfil, "%04x", Old_pc);
  else
    fprintf (Lstfil, "    ");

  for (i = 0; i < P_total && i < 6; i++) fprintf (Lstfil, " %02x", lobyte (P_bytes[i]));

  for (; i < 6; i++) fprintf (Lstfil, "   ");
  fprintf (Lstfil, "  ");

  if (Cflag) {
    if (Cycles)
      fprintf (Lstfil, "[%2d ] ", Cycles);
    else
      fprintf (Lstfil, "      ");
  }

  Line[strlen (Line) - 1] = EOS;      /* No \n */
  fprintf (Lstfil, "%s", Line);       /* just echo the line back out */

  for (; i < P_total; i++) {
    if (i % 6 == 0) fprintf (Lstfil, "\n    ");
    fprintf (Lstfil, " %02x", lobyte (P_bytes[i]));
  }
  fprintf (Lstfil, "\n");
}

/* crc32.c -- compute the CRC-32 of a data stream
 * Copyright (C) 1995-1998 Mark Adler
 * For conditions of distribution and use, see copyright notice in zlib.h 
 */


/* ========================================================================
 * Table of CRC-32's of all single-byte values (made by make_crc_table)
 */
static const unsigned long crc_table[256] = {
  0x00000000L, 0x77073096L, 0xee0e612cL, 0x990951baL, 0x076dc419L,
  0x706af48fL, 0xe963a535L, 0x9e6495a3L, 0x0edb8832L, 0x79dcb8a4L,
  0xe0d5e91eL, 0x97d2d988L, 0x09b64c2bL, 0x7eb17cbdL, 0xe7b82d07L,
  0x90bf1d91L, 0x1db71064L, 0x6ab020f2L, 0xf3b97148L, 0x84be41deL,
  0x1adad47dL, 0x6ddde4ebL, 0xf4d4b551L, 0x83d385c7L, 0x136c9856L,
  0x646ba8c0L, 0xfd62f97aL, 0x8a65c9ecL, 0x14015c4fL, 0x63066cd9L,
  0xfa0f3d63L, 0x8d080df5L, 0x3b6e20c8L, 0x4c69105eL, 0xd56041e4L,
  0xa2677172L, 0x3c03e4d1L, 0x4b04d447L, 0xd20d85fdL, 0xa50ab56bL,
  0x35b5a8faL, 0x42b2986cL, 0xdbbbc9d6L, 0xacbcf940L, 0x32d86ce3L,
  0x45df5c75L, 0xdcd60dcfL, 0xabd13d59L, 0x26d930acL, 0x51de003aL,
  0xc8d75180L, 0xbfd06116L, 0x21b4f4b5L, 0x56b3c423L, 0xcfba9599L,
  0xb8bda50fL, 0x2802b89eL, 0x5f058808L, 0xc60cd9b2L, 0xb10be924L,
  0x2f6f7c87L, 0x58684c11L, 0xc1611dabL, 0xb6662d3dL, 0x76dc4190L,
  0x01db7106L, 0x98d220bcL, 0xefd5102aL, 0x71b18589L, 0x06b6b51fL,
  0x9fbfe4a5L, 0xe8b8d433L, 0x7807c9a2L, 0x0f00f934L, 0x9609a88eL,
  0xe10e9818L, 0x7f6a0dbbL, 0x086d3d2dL, 0x91646c97L, 0xe6635c01L,
  0x6b6b51f4L, 0x1c6c6162L, 0x856530d8L, 0xf262004eL, 0x6c0695edL,
  0x1b01a57bL, 0x8208f4c1L, 0xf50fc457L, 0x65b0d9c6L, 0x12b7e950L,
  0x8bbeb8eaL, 0xfcb9887cL, 0x62dd1ddfL, 0x15da2d49L, 0x8cd37cf3L,
  0xfbd44c65L, 0x4db26158L, 0x3ab551ceL, 0xa3bc0074L, 0xd4bb30e2L,
  0x4adfa541L, 0x3dd895d7L, 0xa4d1c46dL, 0xd3d6f4fbL, 0x4369e96aL,
  0x346ed9fcL, 0xad678846L, 0xda60b8d0L, 0x44042d73L, 0x33031de5L,
  0xaa0a4c5fL, 0xdd0d7cc9L, 0x5005713cL, 0x270241aaL, 0xbe0b1010L,
  0xc90c2086L, 0x5768b525L, 0x206f85b3L, 0xb966d409L, 0xce61e49fL,
  0x5edef90eL, 0x29d9c998L, 0xb0d09822L, 0xc7d7a8b4L, 0x59b33d17L,
  0x2eb40d81L, 0xb7bd5c3bL, 0xc0ba6cadL, 0xedb88320L, 0x9abfb3b6L,
  0x03b6e20cL, 0x74b1d29aL, 0xead54739L, 0x9dd277afL, 0x04db2615L,
  0x73dc1683L, 0xe3630b12L, 0x94643b84L, 0x0d6d6a3eL, 0x7a6a5aa8L,
  0xe40ecf0bL, 0x9309ff9dL, 0x0a00ae27L, 0x7d079eb1L, 0xf00f9344L,
  0x8708a3d2L, 0x1e01f268L, 0x6906c2feL, 0xf762575dL, 0x806567cbL,
  0x196c3671L, 0x6e6b06e7L, 0xfed41b76L, 0x89d32be0L, 0x10da7a5aL,
  0x67dd4accL, 0xf9b9df6fL, 0x8ebeeff9L, 0x17b7be43L, 0x60b08ed5L,
  0xd6d6a3e8L, 0xa1d1937eL, 0x38d8c2c4L, 0x4fdff252L, 0xd1bb67f1L,
  0xa6bc5767L, 0x3fb506ddL, 0x48b2364bL, 0xd80d2bdaL, 0xaf0a1b4cL,
  0x36034af6L, 0x41047a60L, 0xdf60efc3L, 0xa867df55L, 0x316e8eefL,
  0x4669be79L, 0xcb61b38cL, 0xbc66831aL, 0x256fd2a0L, 0x5268e236L,
  0xcc0c7795L, 0xbb0b4703L, 0x220216b9L, 0x5505262fL, 0xc5ba3bbeL,
  0xb2bd0b28L, 0x2bb45a92L, 0x5cb36a04L, 0xc2d7ffa7L, 0xb5d0cf31L,
  0x2cd99e8bL, 0x5bdeae1dL, 0x9b64c2b0L, 0xec63f226L, 0x756aa39cL,
  0x026d930aL, 0x9c0906a9L, 0xeb0e363fL, 0x72076785L, 0x05005713L,
  0x95bf4a82L, 0xe2b87a14L, 0x7bb12baeL, 0x0cb61b38L, 0x92d28e9bL,
  0xe5d5be0dL, 0x7cdcefb7L, 0x0bdbdf21L, 0x86d3d2d4L, 0xf1d4e242L,
  0x68ddb3f8L, 0x1fda836eL, 0x81be16cdL, 0xf6b9265bL, 0x6fb077e1L,
  0x18b74777L, 0x88085ae6L, 0xff0f6a70L, 0x66063bcaL, 0x11010b5cL,
  0x8f659effL, 0xf862ae69L, 0x616bffd3L, 0x166ccf45L, 0xa00ae278L,
  0xd70dd2eeL, 0x4e048354L, 0x3903b3c2L, 0xa7672661L, 0xd06016f7L,
  0x4969474dL, 0x3e6e77dbL, 0xaed16a4aL, 0xd9d65adcL, 0x40df0b66L,
  0x37d83bf0L, 0xa9bcae53L, 0xdebb9ec5L, 0x47b2cf7fL, 0x30b5ffe9L,
  0xbdbdf21cL, 0xcabac28aL, 0x53b39330L, 0x24b4a3a6L, 0xbad03605L,
  0xcdd70693L, 0x54de5729L, 0x23d967bfL, 0xb3667a2eL, 0xc4614ab8L,
  0x5d681b02L, 0x2a6f2b94L, 0xb40bbe37L, 0xc30c8ea1L, 0x5a05df1bL,
  0x2d02ef8dL
};

/* ========================================================================= */
#define DO1(buf, len)  do { crc = crc_table[((int)crc ^ (*buf++)) & 0xff] ^ (crc >> 8); len -= 1; } while (0)
#define DO2(buf, len)  do { DO1(buf, len); DO1(buf, len); } while (0)
#define DO4(buf, len)  do { DO2(buf, len); DO2(buf, len); } while (0)
#define DO8(buf, len)  do { DO4(buf, len); DO4(buf, len); } while (0)

/* ========================================================================= */
unsigned long crc32(unsigned char *buf, unsigned int len)
{
  unsigned long crc = 0UL;
  if (!buf) return 0L;
  crc ^= 0xFFFFFFFFL;
  while (len >= 8) DO8(buf, len);
  if (len >= 4) DO4(buf, len);
  if (len >= 2) DO2(buf, len);
  if (len == 1) DO1(buf, len);
  return crc ^ 0xFFFFFFFFL;
}
#undef DO1
#undef DO2
#undef DO4
#undef DO8


/*
 *      head --- is str2 the head of str1?
 */
int head (char *str1, char *str2) {
  int l = strlen (str2);

  return !strncasecmp (str1, str2, l) && ((str1[l] == EOS) || any (str1[l], " \t\n,+-];*"));
}

/* If compiling with gcc, use the inline versions in the header file instead. */
#ifndef __GNUC__

/* I've replaced most of the original code with more portable current library calls */

/*
 *      any --- does str contain c?
 */
static int any (char c, char *str) {
  if (c == EOS) return NO;
  return strchr(str, c) != NULL;
}

/*
 *      mapdn --- convert A-Z to a-z
 */
static char mapdn (char c) {
  if (isalpha(c) && isupper(c)) return tolower(c);
  return (c);
}

/*
 *      lobyte --- return low byte of an int
 */
static int lobyte (int i) {
  return (i & 0xFF);
}

/*
 *      hibyte --- return high byte of an int
 */
static int hibyte (int i) {
  return ((i >> 8) & 0xFF);
}

/*
 *      alpha --- is character a legal letter
 */
static int alpha (char c) {
  return isalpha(c) || (c == '_') || (c == '.');
}

/*
 *      alphan --- is character a legal letter or digit
 */
static int alphan (char c) {
  return isalnum(c) || (c == '_') || (c == '.') || (c == '$');
}

/*
 *      white --- is character whitespace?
 */
static int white (char c) {
  /* might later switch to either isblank() or strtok(), after checking for unintended consequences... */
  return any (c, " \n\t\r");
}

#endif /* __GNUC__ */
